from dios.lib import *
from dios.options import *
import pandas as pd
import numpy as np
import operator as op
import datetime as dt

from collections import OrderedDict
from pandas.core.dtypes.common import (
    is_list_like,
    is_iterator,
    is_scalar,
    is_number,
    is_integer,
)
from pandas.core.indexing import need_slice


class DictOfSeries:
    """
    DictionaryOfSeries is a collection of pd.Series's which aim to be as close as possible similar to
    pd.DataFrame. The advantage over pd.DataFrame is, that every `column` has its own row-index,
    unlike the former, which provide a single row-index for all columns. This solves problems with
    unaligned pd.Series.

    Dataframe vs DictionaryOfSeries
    -------------------------------

    pd.DataFrame :
      - ``df['a']``      indexing with single label, work on columns, return series
      - ``df[iterabel]`` indexing with list, work on columns, return df
      - ``df[slice]``    indexing with slice, work on rows, return df
      - ``df.loc[..]``   loc[cation] based indexing, work on row and/or columns `labels`, return df
      - ``df.iloc[..]``  i[nteger]loc[cation] based indexing, work on row and/or columns, return df

    DictionaryOfSeries:
      - ``dios['a']``       indexing with single label, work on columns, return series
      - ``dios[iterable]``  indexing with list, work on columns, return dios
      - ``dios[slice]``     indexing with slice, work on (all) index'es, return dios
      - ``dios.loc[..]``    loc[cation] based indexing, work on row and/or columns `labels`, return dios
      - ``dios.iloc[..]``   i[nteger]loc[cation] based indexing, work on row and/or columns, return dios

    Todos:
    -----
    todo: allow any hashable obj as column identifier
         Currently we only allow strings as identifier, to be more df-like we should allow any
         hashable object (unlike df we may should exclude stuff like: ``None`` or ``np.nan`` ??)

    """

    def __init__(self, indextype=None, **kwargs):
        self._data = OrderedDict()

        # We need to keep track of the type of the index of every new Series.
        # If the types differ slicing will almost always fail, because a datetime-like
        # slice cannont work on a numeric index and vice versa..
        if indextype is not None:
            indextype = get_indextype(indextype)
            check_allowed_indextypes(indextype)
        self._indextype = indextype

        # fill initial given values in the dios
        for kw in kwargs:
            self[kw] = kwargs[kw]

    @property
    def columns(self):
        return list(self._data.keys())

    @columns.setter
    def columns(self, new):
        if not isinstance(new, list):
            raise NotImplementedError("Only lists supported so far")

        if len(set(new)) != len(new):
            raise ValueError("Names must be unique")

        if len(new) != len(self.columns):
            raise ValueError(f"Length mismatch: Columns has {len(self.columns)} elements, "
                             f"new values have {len(new)} elements")

        # to keep order, we iterate over self instead of new
        _d = OrderedDict()
        for i, k in enumerate(self.columns):
            _d[new[i]] = self[k]
        self._data = _d

    @property
    def indextype(self):
        return self._indextype

    def _set_indextype(self, idx):
        """ Set indextype of dios.

        Note: If ``self._indextype`` and ``idx`` are of the same type,
        ``self._indextype`` stays unchanged.
        """
        idx = get_indextype(idx)
        check_allowed_indextypes(idx)

        if self._indextype is None:
            self._indextype = idx
        elif self._indextype != idx:
            if dios_options[Options.allow_mixed_indextypes]:
                self._indextype = IdxTypes.mixed
            else:
                raise ValueError(f"Only objects which have a index of type `{self._indextype}` can be inserted.")

    def _check_keys(self, keys):
        missing = [k for k in keys if k not in self.columns]
        if missing:
            raise KeyError(f"{missing} not in index")

    def __getitem__(self, key):
        """
        Get items:
          - ``dios['a']``      -> return a pd.Series
          - ``dios[iterable]`` -> return a DictOfSeries of all given columns [1]
          - ``dios[slice]``    -> return a sliced DictOfSeries of all(!) columns in the dios

        Notes:
          - [1] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """

        # prepare
        if is_iterator(key):
            key = list(key)

        if isinstance(key, str):
            self._check_keys([key])
            return self._get_item(key)

        if isinstance(key, slice):
            return self._slice(self.columns, key)

        if is_list_like(key):
            self._check_keys(key)
            return self._getitem_listlike(key)

        raise KeyError(f"{key}")

    def _get_item(self, key):
        # return always a pd.Series
        return self._data[key]

    def _getitem_listlike(self, keys):
        new = DictOfSeries()
        for k in keys:
            new[k] = self._get_item(k)
        return new

    def _slice(self, keys, slicer):
        """ Return a slice of self"""
        new = DictOfSeries()
        for k in keys:
            new[k] = self._get_item(k)[slicer]
        return new

    def __setitem__(self, key, value):
        """
        Set items:
          - ``dios['a'] = pd.Series()`` -> Set a new pd.Series to an existing column or add it as `new column`.
          - ``dios['a'] = Any`` -> Pass ``any`` to the pd.Series in the corresponding existing(!) column. [1],[2]
          - ``dios[iterable] = Any`` -> Pass ``any`` to the pd.Series's in the corresponding set of columns. [1],[2],[3]
          - ``dios[slice] = Any`` -> Pass ``any`` to all(!) sliced pd.Series's from the current dios. [1],[2]

        Notes:
          - [1] The length of ``Any`` must be be equal to the length of every sliced pd.Series.
          - [2] If ``Any`` is a ``DictOfSeries`` the behavior depends on the option ``dios_to_dios_method``
            in the ``options`` dictionary.
          - [3] If ``iterable`` contains any(!) label that does not exist, a KeyError is raised.
        """

        # prepare
        if is_iterator(key):
            key = list(key)

        # determine action by keys
        if isinstance(key, str):
            if key not in self.columns:
                self._setitem_new(key, value)
            else:
                self._setitem(key, value)
            return

        elif isinstance(key, slice):
            keys = self.columns
            slicer = key

        elif is_list_like(key):
            self._check_keys(key)
            keys = key
            slicer = None

        else:
            raise KeyError(f"{key}")

        if isinstance(value, DictOfSeries):
            self._setitem_dios(keys, slicer, value)
        else:
            for k in keys:
                self._setitem(k, value, slicer)

    def _setitem_new(self, key, value, bypass_checks=False):
        v = value

        # if the checks was already done, we skip them here,
        # also the Errormessage wouldn't fully apply.
        if not bypass_checks:
            if isinstance(v, DictOfSeries):
                v = v.squeeze()

            elif isinstance(v, list):
                v = pd.Series(v)  # upcast

            if not isinstance(v, pd.Series):
                raise ValueError(f"Only pd.Series and DictOfSeries (of length 1) can be assigned new")

        self._set_indextype(v.index)
        self._data[key] = v.copy(deep=True)

    def _setitem(self, key, val, sl=None):
        """ Set a value or a set of values to a single(!) key in self k"""

        # series, dios['a'] = series
        if isinstance(val, pd.Series) and sl is None:
            self._setitem_new(key, val, bypass_checks=True)
            return

        sl = sl or slice(None)

        # label, scalar: dios['a'] = 3.9 or
        # slice, scalar: dios[0:3] = 4.0
        if is_scalar(val):
            self._data[key][sl] = val

        # label,  list: dios['a'] = [0.0, 0.3, 0.0]
        # sclice, list: dios[0:3] = [0.0, 0.3, 0.0]
        elif is_list_like(val):
            # ensure same size
            if len(self._data[key]) == len(val):
                self._data[key][sl] = val
            else:
                raise ValueError(f'Length of values does not match length of index for the key {key}')
        else:
            raise ValueError(f"assignments with a values of type {type(val)} are not supported")
        return

    def _setitem_dios(self, keys, slicer, other):
        keys = get_dios_to_dios_keys(keys, other)
        for k in keys:
            self._setitem(k, other[k], slicer)

    @property
    def loc(self):
        return _LocIndexer(self)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        pd_max_rows = pd.get_option('display.max_rows')
        pd.set_option('display.max_rows', dios_options[Options.disp_max_rows])

        def item2str(k):
            kstr = str(self[k]).replace('\n', '\n ')
            return f'{k}:\n {kstr}\n'

        maxrows = dios_options[Options.disp_max_vars]
        s = ''
        head = maxrows // 2
        tail = len(self.columns) - head
        for i, k in enumerate(self.columns):
            if i < head:
                s += item2str(k)
            elif i == head:
                s += ' ... \t\t\t\t ...\n' * 2
            elif i > tail:
                s += item2str(k)
        pd.set_option('display.max_rows', pd_max_rows)
        return s

    def __bool__(self):
        raise ValueError("The truth value of a DictionaryOfSeries is ambiguous. Use a.empty, a.any() or a.all().")

    @property
    def empty(self):
        return all(self._data[c].empty for c in self._data)

    def all(self):
        return all(all(self._data[c]) for c in self._data)

    def any(self):
        return any(any(self._data[c]) for c in self._data)

    def __len__(self):
        return len(self._data)

    def __reversed__(self):
        return reversed(self._data)

    def __contains__(self, item):
        return item in self._data

    def __iter__(self):
        yield from self._data

    def __delitem__(self, key):
        del self._data[key]
        self.__set_mixed_indextype_from_all_keys()

    def __set_mixed_indextype_from_all_keys(self):
        """ If the index-type of dios is ``mixed`` and the index-type of any stored
        Series change, we need to check the index-type of all other Series, to
        validate the dios-wide index-type."""

        if len(self) == 0:
            self._indextype = None
            return

        if len(self) == 1:
            self._indextype = get_indextype(self.squeeze().index)
            return

        # ``mixed`` isn't allowed in general, so we're done
        if not dios_options[Options.allow_mixed_indextypes]:
            return

        # index-type wasn't ``mixed``, so we're done
        if self._indextype != IdxTypes.mixed:
            return

        # check all types
        types = set()
        for k in self._data.keys():
            idx = self._data[k].index
            types.add(get_indextype(idx))

            # If we have at least two different
            # index-types, ``mixed`` still apply.
            if len(types) > 1:
                return

        # index is of a single new type
        self._indextype = types.pop()
        return

    def __copy__(self):
        return self.copy(deep=True)

    def __deepcopy__(self, memo=None):
        return self.copy(deep=True)

    def copy(self, deep=True):
        new = DictOfSeries()
        new._indextype = self.indextype
        # We use `_data` here because all checks have already been done.
        # So this should be much faster, especially because we use the underlying dict for
        # getting and setting the values, instead of ``__setitem__`` and ``__getitem__``.
        # Note: don't use same approach elsewhere, unless you're very sure what you do.
        for k in self._data:
            new._data[k] = self._data[k].copy(deep=deep)
        return new

    def __op1__(self, op):
        new = DictOfSeries()
        for k in self.columns:
            new[k] = op(self[k])
        return new

    def __op2__(self, other, op):
        new = DictOfSeries()
        if isinstance(other, DictOfSeries):
            keys = get_dios_to_dios_keys(self.columns, other)
            for k in keys:
                new[k] = op(self[k], other[k])
        else:
            for k in self.columns:
                new[k] = op(self[k], other)
        return new

    def __neg__(self):
        return self.__op1__(op.neg)

    def __invert__(self):
        return self.__op1__(op.inv)

    def __abs__(self):
        return self.__op1__(op.abs)

    def __lt__(self, other):
        return self.__op2__(other, op.lt)

    def __le__(self, other):
        return self.__op2__(other, op.le)

    def __eq__(self, other):
        return self.__op2__(other, op.eq)

    def __ne__(self, other):
        return self.__op2__(other, op.ne)

    def __ge__(self, other):
        return self.__op2__(other, op.ge)

    def __gt__(self, other):
        return self.__op2__(other, op.gt)

    def __add__(self, other):
        return self.__op2__(other, op.add)

    def __sub__(self, other):
        return self.__op2__(other, op.sub)

    def __mul__(self, other):
        return self.__op2__(other, op.mul)

    def __mod__(self, other):
        return self.__op2__(other, op.mod)

    def __truediv__(self, other):
        return self.__op2__(other, op.truediv)

    def __floordiv__(self, other):
        return self.__op2__(other, op.floordiv)

    def __pow__(self, other):
        return self.__op2__(other, op.pow)

    def __and__(self, other):
        return self.__op2__(other, op.and_)

    def __or__(self, other):
        return self.__op2__(other, op.or_)

    def __xor__(self, other):
        return self.__op2__(other, op.xor)

    def squeeze(self):
        if len(self) == 1:
            return self[self.columns[0]]
        else:
            return self

    def memory_usage(self, index=True, deep=False):
        mem = 0
        for k in self.columns:
            mem += self[k].memory_usage(index=index, deep=deep)
        return mem

    def foreach(self, f, *args, **kwargs):
        self.pipe(f, *args, **kwargs)

    def pipe(self, f, *args, **kwargs):
        """
        Call a pd.Series function on all series in the DictOfSeries

        :param f: pd.Series function to apply on all Series
        :param args: args for the function
        :param kwargs: kwargs for the function
        :return:
        """
        news = pd.Series()
        newd = DictOfSeries()
        need_dios = False
        # return a copy nevertheless, but also run inplace if inplace=True or
        # if the function not has this option, but work inplace.

        inplace = kwargs.get('inplace', False) or kwargs.get('copy', False)

        def getobj(k, inplace):
            if inplace:
                # self[k] may return a copy, so we need to get
                # the item from the dict, by our mother class.
                return self._data[k]
            else:
                return self._data[k].copy()

        for k in self.columns:
            v = f(getobj(k, inplace), *args, **kwargs)
            if inplace:
                continue
            if isinstance(v, pd.Series):
                newd[k] = v
                need_dios = True
            else:
                newd[k] = pd.Series(data=v)
                news[k] = v

        if inplace:
            return None
        if need_dios:
            return newd.squeeze()
        if len(news) == 0:
            return None
        return news.squeeze()


class _LocIndexer:

    def __init__(self, _dios):
        self._dios = _dios
        # short handles
        self._data = _dios._data
        self._check_keys = _dios._check_keys

    def __getitem__(self, key):
        # if we have a tuple, we have rows and columns
        # if not we have only rows and work on all columns
        if isinstance(key, tuple):
            rkey, ckey, *fail = key
            if fail:
                raise KeyError("To many indexers")

            # prepare ckey
            if is_iterator(ckey):
                ckey = list(ckey)

            # determine columns
            if isinstance(ckey, str):
                self._check_keys([ckey])
                cols = [ckey]
            elif isinstance(ckey, slice):
                cols = self._col_slice_to_col_list(ckey)
            elif is_list_like(ckey):
                self._check_keys(ckey)
                cols = ckey
            else:
                raise KeyError(f"Type {type(ckey)} is not supported to select columns.")
        else:
            cols = self._data.keys()
            rkey = key

        # pass the row-key directly to pd.Series.loc[row-key]
        new = DictOfSeries()
        for c in cols:
            new[c] = self._data[c].loc[rkey]
        return new

    def _col_slice_to_col_list(self, cslice):
        """ see here:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-slicing-with-labels
        """
        keys = list(self._data.keys())
        try:
            start = keys.index(cslice.start) if cslice.start is not None else None
            stop = keys.index(cslice.stop) if cslice.stop is not None else None
        except ValueError:
            raise KeyError("The slice start label or the slice stop label is not present in the columns.")
        if not is_integer(cslice.step) and cslice.step > 0:
            raise TypeError("The step parameter of the slice must be positive integer.")
        return keys[slice(start, stop + 1, cslice.step)]


