from dios.lib import IdxTypes
import warnings


class OptionsWarning(UserWarning):
    pass


class OptionsError(RuntimeError):
    pass


class Options:
    """storage class for the keys in ``dios_options``"""

    """Set the number of rows and variables to display in a call that use
    ``__repr__`` or ``__str__`` like e.g. ``print(dios)`` do."""
    disp_max_rows = "disp_max_rows "
    disp_max_vars = "disp_max_vars"

    """
    0: accept all
    1: accept if at least one keys is is in both DioS
    2: accept if all keys of the src-DioS in the dest-DioS
    3: accept if both dios have the exact same keys (makes only sense for assignments with slicer,
      otherwise its the same than creating a new dios)"""
    dios_to_dios_method = "dios_to_dios_method"

    """
    If we have different types of indexes in the dios, slicing will almost always fail.
    It is because, eg. a numeric slice cannot work on a pd.DatetimeIndex and vice versa.
    To set this to True is highly experimental, any arising issues or errors should be
    handled by the user."""
    allow_mixed_indextypes = "allow_mixed_indextypes"

    allowed_indextypes = "allowed_indextypes"


class __OptionsDict(dict):
    """ Simple dict that throw a warning, if a special value is inserted at a special key"""
    def __setitem__(self, key, value):
        # throw a warning when user set ``mixed_indextyes = True``
        if key == Options.allow_mixed_indextypes and value:
            warnings.warn(f"Using ``dios_option[{Options.allow_mixed_indextypes}]=True`` is highly experimental, "
                          f"please do not report any bugs!", OptionsWarning)
        return super().__setitem__(key, value)


# set default values
dios_options = __OptionsDict()
dios_options[Options.disp_max_rows] = 10
dios_options[Options.disp_max_vars] = 4
dios_options[Options.dios_to_dios_method] = 3
dios_options[Options.allow_mixed_indextypes] = False
dios_options[Options.allowed_indextypes] = [IdxTypes.datetime, IdxTypes.nunmeric]


def check_allowed_indextypes(idxtype):
    if idxtype not in dios_options[Options.allowed_indextypes]:
        raise RuntimeError(f"The index type `{idxtype}` is not allowed by the "
                           f"`dios_option[{Options.allowed_indextypes}] = {dios_options[Options.allowed_indextypes]}`")


def get_dios_to_dios_keys(keys, other):
    # we can assume that all keys are exist in self._data
    method = dios_options[Options.dios_to_dios_method]
    err_append = "consider changing dios.option['dios_to_dios_method']"

    # assign where possible, otherwise ignore
    if method == 0:
        keys = [k for k in keys if k in other.columns]

    # at least one key must be in self
    elif method == 1:
        keys = [k for k in keys if k in other.columns]
        if not keys:
            raise KeyError("src-DioS and dest-DioS need to share at least one key, " + err_append)

    # all keys must be in self, but more keys could exist in other,
    # eg. ``dios['a'] = dios[['a','b']]`` will update column-a but not column-b
    # eg. ``dios[['a','b']] = dios['a']`` will fail
    elif method == 2:
        fail = [k for k in keys if k not in other.columns]
        if fail:
            raise KeyError(f"{fail} are missing in the destiny-dios, " + err_append)

    # keys in both dios's must be equal
    elif method == 3:
        fail = set(keys).symmetric_difference(set(other.columns))
        if fail:
            raise KeyError(f"{fail} is not in both of src- and dest-dios, " + err_append)

    else:
        raise OptionsError(f"{method} is an invalid value for dios.option[dios_to_dios]")

    return keys
